BROADCAST_ADDRESS = 0
BROADCAST_ADDRESS_SILENT = 0xFFFFFFFF

NCP_PROTOCOL_ID = 6

NCP_CRC_POLYNOM = 0x8005

CMD_DATA = 6
CMD_ERROR = 7

class NcpException(Exception):
    def __init__(self, text):
        self.text = text
    def __str__(self):
        return self.text

class NcpErrorException(NcpException):
    errors = {
        1: ('ERR_GENERAL', 'Общая ошибка'),
        2: ('ERR_PARAMS', 'Ошибка параметров'),
        3: ('ERR_UNSUPPORTED', 'Неизвестный код'),
        4: ('ERR_WRITE', 'Ошибка записи'),
        5: ('ERR_OUTOFMEMORY', 'Недостаточно памяти'),
        6: ('ERR_WRONGADDR', 'Неверный адрес'),
        7: ('ERR_WRONGDATA', 'Некоректные данные в команде'),
        8: ('ERR_BUSY', 'Устройство занято'),
        9: ('ERR_CONNECT', 'Нет связи')         
    }
    def __init__(self, code, data = ''):
        self.code = code
        self.data = data

    def __str__(self):
        descr = NcpErrorException.errors.get(self.code, ('ERR_{}'.format(self.code), 'Ошибка {}'.format(self.code)))
        return  descr[0] + ': ' + descr[1]

def CRC(buf: bytes):
    crc = 0
    for b in buf:
        crc ^= b << 8
        for _ in range(8):
            if crc & 0x8000: 
                crc = (crc << 1) ^ NCP_CRC_POLYNOM
            else:
                crc = crc << 1
            crc &= 0xFFFF
    return crc

class Ncp:
    def __init__(self):
        pass

    def decode(self, buf: bytes):
        if buf[0] != NCP_PROTOCOL_ID:
            raise NcpException('Wrong NCP protocol id')
        if len(buf) < 9:
            raise NcpException('Packet too small')
        if CRC(buf[:-2]) != int.from_bytes(buf[-2:], byteorder = 'big', signed = False):
            raise NcpException('Wrong CRC')
        if buf[5] & 0x7F != 0:
            raise NcpException('Unknown flags')
        self.mac = int.from_bytes(buf[1:5], byteorder = 'little', signed = False)
        self.ans = buf[5] == 0x80
        self.cmd = buf[6]
        self.data = buf[7:-2]

    def encode(self):
        if type(self.mac) is not int or self.mac < 0 or self.mac > 0xFFFFFFFF:
            raise NcpException('Invalid MAC address')
        if type(self.cmd) is not int or self.cmd < 0 or self.cmd > 0xFF: 
            raise NcpException('Invalid command')

        res = bytes([NCP_PROTOCOL_ID]) + \
            self.mac.to_bytes(4, byteorder = 'little', signed = False) + \
            bytes([0, self.cmd]) + \
            self.data
        res += CRC(res).to_bytes(2, byteorder = 'big', signed = False)
        return res



    def check_error(self):
        if self.cmd == CMD_ERROR:
            raise NcpErrorException(self.data[0], self.data[1:])



from .core import Ncp, NcpException, CMD_DATA

class NcpRequest:
    def __init__(self, mac):
        self.ncp = Ncp()
        self.ncp.mac = mac

    def data(self, data):
        self.ncp.cmd = CMD_DATA
        self.ncp.data = data
        return self.ncp.encode()


from .core import NcpErrorException, BROADCAST_ADDRESS, BROADCAST_ADDRESS_SILENT, CMD_DATA
from .answer import NcpAnswer
from .request import NcpRequest

from . import slip as SLIP

def decode_data(data: bytes, slip: bool = None) -> (int, bytes):
    """Кодирует данные в NCP data

    Args:
        data (bytes): данные

    Returns:
        int, bytes: адрес, закодированный пакет
    """
    if slip is None:
        slip = data[0] == 0xC0
        
    if slip: data = SLIP.decode(data)
    ncp = NcpAnswer(data, command = CMD_DATA)
    return ncp.mac, ncp.data()

def encode_data(data, mac: int = BROADCAST_ADDRESS, slip: bool = False):
    ncp = NcpRequest(mac)
    packet = ncp.data(data)
    if slip: 
        packet = SLIP.encode(packet)
    return packet


from .core import Ncp, NcpException

class NcpAnswer:
    def __init__(self, data: bytes, mac: int = None, command: int = None):
        self.mac = mac
        self.command = command
        self.ncp = Ncp()

        if data is not None:            
            self.decode(data)

    def decode(self, data):
        self.ncp.decode(data)
        if self.mac is not None and self.mac != self.ncp.mac:
            raise NcpException('Wrong answer MAC address')
        self.mac = self.ncp.mac

        self.ncp.check_error()

        if self.command is not None and self.command != self.ncp.cmd:
            raise NcpException('Wrong answer command')


    def data(self):
        return self.ncp.data


from setuptools import setup, find_packages
from os import path
import unittest

HERE = path.abspath(path.dirname(__file__))

with open(path.join(HERE, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


def test_suite():
    test_loader = unittest.TestLoader()
    test_suite = test_loader.discover('tests', pattern='test_*.py')
    return test_suite


setup(
    name='ul-pyncp',
    version='1.0.5',
    description='Python pyncp library',
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT",
    packages=find_packages(),
    platforms='any',
    classifiers=[
                "Intended Audience :: Developers",
                "License :: OSI Approved :: MIT License",
                "Programming Language :: Python",
                "Programming Language :: Python :: 3.6",
                "Programming Language :: Python :: 3.7",
                "Programming Language :: Python :: 3.8",
                "Programming Language :: Python :: 3.9",
                "Operating System :: OS Independent"
            ],
    include_package_data=True,
    test_suite='setup.test_suite',
)


import re
class SlipException(Exception):
    def __str__(self):
        return 'SLIP decode error'

def encode(buf: bytes) -> bytes:
    """Кодирует данные в SLIP

    Args:
        buf (bytes): данные

    Returns:
        bytes: закодированный SLIP пакет
    """
    return b'\xC0' + buf.replace(b'\xDB',b'\xDB\xDD').replace(b'\xC0', b'\xDB\xDC') + b'\xC0'

def decode(buf: bytes) -> bytes:
    """Декодирует данные из SLIP

    Args:
        buf (bytes): SLIP пакет

    Raises:
        SlipException: при неверном формате пакета

    Returns:
        bytes: декодированные данные
    """
    if buf[0] != 0xC0 or buf[-1] != 0xC0: raise SlipException()
    if re.search(b'\xDB[^\xDC\xDD]', buf) is not None: raise SlipException()
    return buf[1:-1].replace(b'\xDB\xDC', b'\xC0').replace(b'\xDB\xDD', b'\xDB')